# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# System
from typing import Optional, List
import os

# Pip
from selenium_browser import AddonUtils, AddonInstallSettings

# -------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------ class: AddonManager ----------------------------------------------------- #

class AddonManager:

    # --------------------------------------------------------- Init --------------------------------------------------------- #

    def __init__(
        self,
        browser, #: Firefox
    ):
        self.browser = browser

        self.__supported_addon_file_extensions = ['xpi', 'zip']
        self.__addon_url_format = 'moz-extension://{}'


    # ---------------------------------------------------- Public methods ---------------------------------------------------- #

    def get_all_addon_settings(
        self,
        addons_folder_path: Optional[str] = None,
        user_addon_settings: Optional[List[AddonInstallSettings]] = None,
        lib_addon_settings: Optional[List[AddonInstallSettings]] = None
    ) -> List[AddonInstallSettings]:
        return AddonUtils.get_all_addons(
            supported_addon_file_extensions=self.__supported_addon_file_extensions,
            addons_folder_path=addons_folder_path,
            user_addon_settings=user_addon_settings,
            lib_addon_settings=lib_addon_settings
        )

    def addons_from_folder(
        self,
        folder_path: str
    ) -> List[AddonInstallSettings]:
        return AddonUtils.addons_from_folder(
            folder_path=folder_path,
            supported_addon_file_extensions=self.__supported_addon_file_extensions
        )

    def install_addons(
        self,
        addons_settings: List[AddonInstallSettings],
        temporary: Optional[bool] = False
    ) -> List[Optional[str]]:
        return [
            self.install_addon(addon_settings=addon_settings, temporary=temporary)
            for addon_settings in addons_settings
        ]

    def install_addon(
        self,
        addon_settings: AddonInstallSettings,
        temporary: Optional[bool] = False,
        debug: bool = False
    ) -> Optional[str]:
        if addon_settings.path.split('.')[-1] not in self.__supported_addon_file_extensions:
            if debug:
                print('Will not install \'{}\'. Reason: not one of the needed formats: {}'.format(
                    addon_settings.path,
                    self.__supported_addon_file_extensions
                ))

            return None

        org_handle = self.browser.driver.current_window_handle
        addon_id = self.browser.driver.install_addon(addon_settings.path, temporary=temporary)

        if not addon_id:
            if debug:
                print('Could not install addon with path \'{}\''.format(addon_settings.path))

                return None

        if debug:
            print('Did install addon with id \'{}\' and path \'{}\''.format(addon_id, addon_settings.path))

        internal_addon_id = self.addon_public_id_to_private_id_function(self.browser, addon_id)
        org_current_url = self.browser.current_url

        def reset(org_handle, org_current_url):
            if len(self.browser.window_handles) > 1:
                for handle in self.browser.window_handles:
                    self.browser.switch_to_window(handle)

                    if internal_addon_id in self.browser.current_url:
                        self.browser.close()

                        return reset(org_handle, org_current_url)

            if org_handle not in self.browser.window_handles:
                org_handle = self.browser.window_handles[0]

            self.browser.switch_to_window(org_handle)
            self.browser.get(org_current_url)

        reset(org_handle, org_current_url)

        addon_settings.post_install_action(
            self.browser,
            addon_id,
            internal_addon_id,
            self.__addon_url_format.format(internal_addon_id)
        )

        reset(org_handle, org_current_url)

        return addon_id

    @staticmethod
    def addon_public_id_to_private_id_function(
        firefox,
        public_addon_id: str
    ) -> str:
        import time
        start_time = time.time()

        while not firefox.prefs.extensions_webextensions_uuids.get(public_addon_id) and time.time() - start_time < 5:
            time.sleep(0.05)

        return firefox.prefs.extensions_webextensions_uuids.get(public_addon_id)


# -------------------------------------------------------------------------------------------------------------------------------- #


# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# System
from typing import Optional, Union, List, Tuple
import os, shutil, time

# Pip
from noraise import noraise
from selenium_browser import Browser, AddonInstallSettings, Utils as BrowserUtils
from kproxy import Proxy

from selenium.webdriver import Firefox as FirefoxWebDriver
from selenium.webdriver.firefox.firefox_binary import FirefoxBinary

import geckodriver_autoinstaller

# Local
from .firefox_addons import AddonManager, FoxyProxyAddonSettings

from .__firefox_properties import FirefoxProperties
from .__constants import Constants
from .__utils import Utils

# -------------------------------------------------------------------------------------------------------------------------------- #



# -------------------------------------------------------- class: Firefox -------------------------------------------------------- #

class Firefox(
    Browser,
    FirefoxProperties
):

    # --------------------------------------------------------- Init --------------------------------------------------------- #

    def __init__(
        self,

        # cookies
        cookies_folder_path: Optional[str] = None,
        cookies_id: Optional[str] = None,
        pickle_cookies: bool = False,

        # proxy
        proxy: Optional[Union[Proxy, str]] = None,
        # proxy - legacy (kept for convenience)
        host: Optional[str] = None,
        port: Optional[int] = None,

        # addons
        addons_folder_path: Optional[str] = None,
        addon_settings: Optional[List[AddonInstallSettings]] = None,
        # addons - legacy (kept for convenience)
        extensions_folder_path: Optional[str] = None,

        # other paths
        geckodriver_path: Optional[str] = None,
        firefox_binary_path: Optional[str] = None,
        profile_path: Optional[str] = None,

        # profile settings
        private: bool = False,
        full_screen: bool = True,
        language: str = 'en-us',
        user_agent: Optional[str] = None,
        disable_images: bool = False,
        mute_audio: bool = False,

        # firefox option settings
        screen_size: Optional[Tuple[int, int]] = None, # (width, height)
        headless: bool = False,
        home_page_url: Optional[str] = None,

        # selenium-wire support
        webdriver_class: Optional = None,

        # find function
        default_find_func_timeout: int = 2.5
    ):
        '''EITHER PROVIDE 'cookies_id' OR  'cookies_folder_path'.
           IF 'cookies_folder_path' is None, 'cokies_id', will be used to calculate 'cookies_folder_path'
           IF 'cokies_id' is None, the name of the 'profile_path' follder wil lbe used. if that is Nonne too, 'test' will be used

           webdriver_class: override class used to create webdriver (for example: seleniumwire.webdriver.Firefox), Defaults to: 'selenium.webdriver.Firefox'
        '''

        self.source_profile_path = profile_path

        cookies_folder_path = BrowserUtils.cookies_folder_path(cookies_folder_path, cookies_id, profile_path)
        os.makedirs(cookies_folder_path, exist_ok=True)

        user_agent = BrowserUtils.user_agent(user_agent, BrowserUtils.user_agent_path(cookies_folder_path, cookies_id, profile_path))

        proxy = Utils.proxy(
            proxy=proxy,
            host=host,
            port=port
        )

        addon_settings = addon_settings or []

        if proxy and proxy.needs_auth:
            addon_settings.append(FoxyProxyAddonSettings())

        super().__init__(
            webdriver_class or FirefoxWebDriver,
            cookies_folder_path=cookies_folder_path,
            cookies_id=cookies_id,
            pickle_cookies=pickle_cookies,
            proxy=proxy,
            default_find_func_timeout=default_find_func_timeout,
            webdriver_executable_path=Utils.get_geckodriver_path(geckodriver_path),
            firefox_profile=Utils.profile(
                user_agent=BrowserUtils.user_agent(
                    user_agent=user_agent,
                    file_path=BrowserUtils.user_agent_path(
                        cookies_folder_path=cookies_folder_path,
                        cookies_id=cookies_id,
                        profile_path=profile_path
                    )
                ),
                language=language,
                private=private,
                disable_images=disable_images,
                mute_audio=mute_audio,
                proxy=proxy if proxy and not proxy.needs_auth else None,
                path=profile_path
            ),
            firefox_options=Utils.options(
                screen_size=screen_size,
                headless=headless,
                home_page_url=home_page_url
            ),
            firefox_binary=Utils.get_firefox_binary(firefox_binary_path)
        )

        if full_screen:
            self.driver.fullscreen_window()

        am = AddonManager(self)

        am.install_addons(
            addons_settings=am.get_all_addon_settings(
                addons_folder_path=addons_folder_path,
                user_addon_settings=addon_settings
            ),
            temporary=False
        )


    # ---------------------------------------------------- Public methods ---------------------------------------------------- #

    @noraise(default_return_value=False)
    def backup_profile(
        self,
        target_profile_path: Optional[str] = None,
        delete_cache: bool = True,
        delete_storage: bool = True,
        max_copy_tries: int = 5
    ) -> bool:
        target_profile_path = target_profile_path or self.source_profile_path

        if os.path.exists(target_profile_path):
            shutil.rmtree(target_profile_path)

        current_copy_try = 0

        while not os.path.exists(target_profile_path) and current_copy_try < max_copy_tries:
            try:
                shutil.copytree(self.temp_profile_folder_path, target_profile_path)
            except:
                pass

            time.sleep(0.1)

        to_remove = [v for v in [
            'cache2' if delete_cache else None,
            'storage' if delete_storage else None,
            'storage.sqlite' if delete_storage else None
        ] if v]

        for v in to_remove:
            p = os.path.join(target_profile_path, v)

            if os.path.exists(p):
                shutil.rmtree(p) if os.path.isdir(p) else os.remove(p)

        return os.path.exists(target_profile_path)


# -------------------------------------------------------------------------------------------------------------------------------- #


# ------------------------------------------------------------ Imports ----------------------------------------------------------- #

# System
from typing import Optional
import os, json

# Pip
from selenium_browser import BrowserProperties
from noraise import noraise

from selenium.webdriver.firefox.webdriver import WebDriver

# Local
from .models import FirefoxCapabilities, Prefs
from .__constants import Constants
from .__utils import Utils

# -------------------------------------------------------------------------------------------------------------------------------- #



# --------------------------------------------------- class: FirefoxProperties --------------------------------------------------- #

class FirefoxProperties(BrowserProperties):

    # --------------------------------------------------- Public properties -------------------------------------------------- #

    driver: WebDriver

    @property
    def capabilities(self) -> FirefoxCapabilities:
        return FirefoxCapabilities(self.driver.capabilities)

    @property
    def prefs(self) -> Optional[Prefs]:
        @noraise()
        def _prefs() -> Optional[Prefs]:
            return Prefs(Utils.parse_prefs_js(open(os.path.join(self.capabilities.moz_profile, Constants.PREFS_JS_NAME), 'r').read()))

        return _prefs()

    @property
    def temp_profile_folder_path(self) -> Optional[str]:
        @noraise(print_exc=False)
        def _temp_profile_folder_path() -> Optional[str]:
            return self.capabilities.moz_profile

        return _temp_profile_folder_path()


# -------------------------------------------------------------------------------------------------------------------------------- #